/******************************  main.c  *******************************

  Purpose:	Main routine, driving the MPHF creation.

  Provenance:	Written and tested by Q. Chen and E. Fox, March 1991.
  		Edited and tested by S. Wartik, April 1991.

  Notes:	When compiled, the resulting program is used as follows:

  			phf I L R O

		where:

	            I	Name of the file to be used as input.  It should contain
		    	one or more newline-terminated strings.

		    L	The number of lines in I.

		    R	A real number, giving a ratio between L and the size of
		    	the hashing function generated. 1.0 is usually a viable
			value.  In general, L*R should be an integer.

		    O	Name of a file to be used as output.  It will contain
		    	the MPHF if one is found.
**/

#include <stdio.h>
#include <math.h>
#include <string.h>

#include "types.h"
#include "support.h"

#ifdef __STDC__

extern void	ordering( arcsType *arcs, verticesType *vertices );

extern void	allocate_arcs( arcsType* arcs, int n );
extern void	allocate_vertices( verticesType* vertices, int n );
extern void	free_arcs( arcsType* arcs );
extern void	free_vertices( verticesType* vertices );

extern void	exit();

#else

extern void	ordering();

extern void	allocate_arcs();
extern void	allocate_vertices();
extern void	free_arcs();
extern void	free_vertices();

extern void	exit();

#endif

/*************************************************************************
	main(argc, argv)

  Returns:	int -- zero on success, non-zero on failure.

  Purpose:	Take the inputs and call three routines to carry out mapping,
		ordering and searching three tasks.  If they all succeed,
		write the MPHF to the spec file.
**/

main( argc, argv )
	int	argc;
	char	*argv[];	/* arg1: key file;    arg2: key set size;  */ 
				/* arg3: ratio;       arg4: spec file      */
{
    int		status,         /* Return status variable.                   */
		seed;           /* Seed used to initialize the three random  */
				/* tables.				     */
    char	*key_file_name,
		*specification_file_name;

    int		lines_in_keyword_file;
    double	ratio;
 
    arcsType	    arcs;	/* These variables hold all the arcs	*/
    verticesType    vertices;   /* and vertices generated.		*/

    if ( argc != 5 ) {
        fprintf(stderr, "Usage: %s keywords kw-lines ratio output-file\n",
		argv[0]);
	exit(1);
    }
    
    key_file_name = argv[1];
    if ( (lines_in_keyword_file = atoi(argv[2])) <= 0 ) {
	fputs("The 2nd parameter must be a positive integer.\n", stderr);
	exit(1);
    }
    else if ( (ratio = atof(argv[3])) <= 0.0 ) {
	fputs("The 3rd parameter must be a positive floating-point value.\n",
	      stderr);
	exit(1);
    }

    allocate_arcs( &arcs, lines_in_keyword_file );
    allocate_vertices( &vertices, (int)(lines_in_keyword_file * ratio) );
    specification_file_name = argv[4];

    if ( (status = mapping( key_file_name, &arcs, &vertices, &seed )) == NORM ) {
	ordering( &arcs, &vertices );
	if ( (status = searching( &arcs, &vertices )) == NORM &&
	     (status = verify_mphf( &arcs, &vertices )) == NORM )
	    write_gfun( &arcs, &vertices, seed, specification_file_name );
    }

    free_arcs( &arcs );
    free_vertices( &vertices );
    fprintf(stderr, "MPHF creation %s.\n",
		    (status == NORM ? "succeeded" : "failed"));

    return(status);
}



/*************************************************************************

       allocate_arcs( arcsType*, int )

  Returns:	void       

  Purpose:	Given an expected number of arcs, allocate space for an arc data
		structure containing that many arcs, and place the number of
		arcs in the "no_arcs" field of the arc data structure.
**/

void allocate_arcs( arcs, n )
    arcsType	*arcs;		/* out: Receives allocated storage.	*/
    int		n;		/* in: Expected number of arcs.		*/
{
    arcs->no_arcs = n;
    arcs->arcArray = (arcType*) owncalloc( sizeof(arcType), n );
}



/*************************************************************************

	allocate_vertices( verticesType* , int )

  Purpose:	Given an expected number of vertices, allocate space for a vertex
		data structure containing that many vertices, and place the
		number of vertices in the "no_vertices" field of the vertex
		data structure.
**/

void allocate_vertices( vertices, n )
    verticesType	*vertices;	/* out: Receives allocated storage. */
    int			n;		/* in: Expected number of vertices. */
{
    if (n % 2 != 0) n++;
    vertices->no_vertices = n;
    vertices->vertexArray = (vertexType*) owncalloc(sizeof(vertexType), n );
}



/*************************************************************************

	free_arcs( arcsType* )

  Purpose:	Deallocate space for an arc data structure.
**/

void free_arcs( arcs )
    arcsType	*arcs;			/* in out: Space to de-allocate.  */
{
    free( (char *)arcs->arcArray );
}


/*************************************************************************

	free_vertices( verticesType* )

  Purpose:	Deallocate space for a vertex data structure.
**/

void free_vertices( vertices )
    verticesType	*vertices;	/* in out: Space to de-allocate.  */
{
    free( (char *)vertices->vertexArray );
}
