/*********************************  map.c  **********************************

  Purpose:	Implement the mapping stage of the MPHF algorithm.

  Provenance:	Written and tested by Q. Chen and E. Fox, March 1991.
  		Edited and tested by S. Wartik, April 1991.

  Notes:	To save space, the "g" field of the vertex structure is used
  		here to hold a vertex's degree.

**/


#include <stdio.h>

#include "types.h"
#include "pmrandom.h"
#include "rantab.h"
#include "comphfns.h"

#ifdef __STDC__

extern void	initialize_arcs( arcsType *arcs );
extern void	initialize_vertices( verticesType *vertices );
extern int	check_dup( arcType *firstArc );
extern int	construct_graph( arcsType *arcs, verticesType *vertices );
extern void	map_to_triples( char *key_file, arcsType *arcs,
				int r, randomTablesType tables );

extern void	exit( int status );

#else

extern void	initialize_arcs();
extern void	initialize_vertices();
extern int	check_dup();
extern int	construct_graph();
extern void	map_to_triples();

extern void	exit();

#endif


/*************************************************************************

	mapping( char*, arcsType*, verticesType*, int )

  Return:	int -- NORM if a mapping can be found, ABNORM if not.
  
  Purpose:	Perform the mapping stage: Map all keys to triples and construct
  		the bipartite graph.  This involves:

		  --	Allocating the arcs and vertices structures.
		  --	Generating the h0, h1, and h2 functions.
		  --	Building the lists of edges, ordered by degree.

**/

int mapping( key_file, arcs, vertices, seed )
    char	*key_file;	/* in: name of file containing keys.	*/
    arcsType	*arcs;          /* out: arcs in bipartite graph.	*/
    verticesType *vertices; 	/* out: vertices in bipartite graph.	*/
    int		*seed;          /* out: seed selected to initialize the */
				/* random tables.			*/
{
    int	mapping_tries = 0;
    randomTablesType randomTables;      /* Three random number tables.	     */

    while ( mapping_tries++ < MAPPINGS ) {
	initialize_arcs( arcs );
	initialize_vertices( vertices );
	initialize_randomTable( randomTables, seed );
	map_to_triples( key_file, arcs, vertices->no_vertices/2, randomTables );
	if ( construct_graph(arcs, vertices) == NORM )
	    return(NORM);
	fputs((mapping_tries < MAPPINGS ? "Trying again.\n" : "Giving up.\n"),
	      stderr);
    }
    return(ABNORM);
}


/*************************************************************************

	map_to_triples( char*, arcsType*, int, randomTablesType )

  Return:	void
  
  Purpose:	Compute triples of (h0, h1, h2) for all keys and store
		them in the arc data structure.
**/

void map_to_triples(key_file, arcs, r, tables)
    char	*key_file;      /* in: key file name            */
    arcsType	*arcs;          /* out: the arcs data structure */
    int		r;              /* in: size of h1 or h2 side    */
    randomTablesType tables;	/* in: random number tables     */
{
    FILE	*fp;                  /* Input file pointer.	*/
    int		i = 0;                /* Iterator over arcs.	*/
    char	string[MAX_KEY_LENG]; /* Key string holder.	*/

    if ( (fp = fopen(key_file, "r")) == NULL ) {
	fprintf(stderr, "Can't read \"%s\".\n", key_file);
	exit(1);
    }

    while ( fgets( string, MAX_KEY_LENG, fp ) != NULL &&
	   i < arcs->no_arcs ) {
	string[strlen(string)-1] = '\0';      /* Exclude the '\n'. */
	compute_h012( arcs->no_arcs, r, tables, string, &arcs->arcArray[i++] );
    }

    if ( i != arcs->no_arcs ) {
	fprintf(stderr, "File \"%s\" contains %d keys, not %d. ",
		key_file, i, arcs->no_arcs);
	fputs("Re-execute with correct value.\n", stderr);
	exit(1);
    }
    else if ( ! feof(fp ) ) {
	fprintf(stderr, "File \"%s\" contains more than %d keys. ",
		key_file, arcs->no_arcs);
	fputs("Re-execute with correct value.\n", stderr);
	exit(1);
    }

    fclose(fp);
}



/*************************************************************************

	construct_graph( arcsType*, verticesType* )

  Return:	int -- NORM if a graph can be built without duplicate arcs,
  		ABNORM if it can't.

  Purpose:	Construct the bipartite graph out of triples.  On successful
  		return,

		    --	Each vertex's degree has been determined, and placed
		    	in its "g" field.
		    --	The maximal degree of the graph has been determined.
		    --	The "first_edge" field of vertices is a linked list
		    	of adjacent edges.
**/

int construct_graph( arcs, vertices )
    arcsType	 *arcs;		/* in out: arcs.			*/
    verticesType *vertices;	/* in out: vertices.			*/
{
    int i,             /* Iterator over all arcs.			*/
	j,             /* j = 0 and 1 for h1 and h2 side, respectively.	*/
	status = NORM,
	vertex;

    for ( j = 0; j < 2; j++ ) {
	for ( i = 0; i < arcs->no_arcs; i++ ) {
	    vertex = arcs->arcArray[i].h12[j];  /* Update vertex degree */
	    vertices->vertexArray[vertex].g++;  /* count and vertex     */
	    arcs->arcArray[i].next_edge[j] =    /* adjacency list.      */
		    vertices->vertexArray[vertex].first_edge;
	    vertices->vertexArray[vertex].first_edge = &arcs ->arcArray[i];

	    if ( (j == 0) && check_dup( &arcs->arcArray[i] ) == ABNORM ) {
		fputs("Duplicate found.\n", stderr);
		status = ABNORM;
		break;
	    }
		     /* Figure out the maximal degree of the graph. */
	    if ( vertices->vertexArray[vertex].g > vertices->maxDegree )
		vertices->maxDegree = vertices->vertexArray[vertex].g;
	}
    }
    return(status);
}

/*************************************************************************

	check_dup( arcType )

  Return:	int -- NORM if no duplicate triple exists, ABNORM if one does.

  Purpose:	Test if some arc on the arc list has an identical triple to
		the first arc on the list.
**/

int check_dup(firstArc)
    arcType *firstArc;	/* in: arc at the head of a list.	*/
{
    arcType *arc = firstArc->next_edge[0];
    while ( arc != 0 ) {
	if ( ( firstArc->h0 == arc->h0 ) &&
	    ( firstArc->h12[1] == arc->h12[1] ) )
	    return(ABNORM);                            /* Duplication found. */
	arc = arc->next_edge[0];
    }
    return(NORM);                                      /* No duplication.    */
}



/*************************************************************************

	initialize_arcs( arcsType* )

  Return:	void
  
  Purpose:	Make the edge pointers of each arc nil.
**/

void initialize_arcs( arcs )
    arcsType *arcs;	/* out: arcs structure.			*/
{
    int i;

    for ( i = 0; i < arcs->no_arcs; i++ ) {
	arcs->arcArray[i].next_edge[0] = 0;
	arcs->arcArray[i].next_edge[1] = 0;
    }
}

/*************************************************************************

	initialize_vertices( verticesType* )

  Return:	void
  
  Purpose:	For each vertex, set the degree to 0 and make the
  		edge list empty.
**/

void initialize_vertices( vertices )
    verticesType *vertices;	/* out: vertex structure.	*/
{
    int i;

    vertices->maxDegree = 0;
    for ( i = 0; i < vertices->no_vertices; i++ ) {
	vertices->vertexArray[i].first_edge = 0;
	vertices->vertexArray[i].g = 0;
    }
}
